@@ -1,6 +1,6 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2009 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
  *
  * This product is released under the terms of the General Public Licence,
  * version 2 (GPLv2). Please refer to the file LICENSE (included with this
@@ -12,14 +12,15 @@
  * distribution.
  *
  * If any of the files related to licensing are missing or if you have any
- * other questions related to licensing please contact Breach Security, Inc.
- * directly using the email address support@breach.com.
+ * other questions related to licensing please contact Trustwave Holdings, Inc.
+ * directly using the email address support@trustwave.com.
  *
  */
 #include <limits.h>
 
 #include "http_core.h"
 #include "http_request.h"
+#include "http_connection.h"
 
 #include "modsecurity.h"
 #include "apache2.h"
@@ -29,6 +30,8 @@
 #include "msc_logging.h"
 #include "msc_util.h"
 
+#include "ap_mpm.h"
+#include "scoreboard.h"
 
 /* ModSecurity structure */
 
@@ -50,6 +53,18 @@ apr_file_t DSOLOCAL *guardianlog_fd = NULL;
 
 char DSOLOCAL *guardianlog_condition = NULL;
 
+unsigned long int DSOLOCAL msc_pcre_match_limit = 0;
+
+unsigned long int DSOLOCAL msc_pcre_match_limit_recursion = 0;
+
+unsigned long int DSOLOCAL conn_read_state_limit = 0;
+
+static int server_limit, thread_limit;
+
+typedef struct {
+    int child_num;
+    int thread_num;
+} sb_handle;
 
 /* -- Miscellaneous functions -- */
 
@@ -225,9 +240,24 @@ int perform_interception(modsec_rec *msr) {
             break;
     }
 
+    /* If the level is not high enough to add an alert message, but "auditlog"
+     * is enabled, then still add the message. */
+    if ((log_level > 3) && (actionset->auditlog != 0)) {
+        *(const char **)apr_array_push(msr->alerts) = msc_alert_message(msr, actionset, NULL, message);
+    }
+
     /* Log the message now. */
     msc_alert(msr, log_level, actionset, message, msr->intercept_message);
 
+    /* However, this will mark the txn relevant again if it is <= 3,
+     * which will mess up noauditlog.  We need to compensate for this
+     * so that we do not increment twice when auditlog is enabled and
+     * prevent incrementing when auditlog is disabled.
+     */
+    if ((actionset->auditlog == 0) && (log_level <= 3)) {
+        msr->is_relevant--;
+    }
+
     return status;
 }
 
@@ -1042,6 +1072,60 @@ static void modsec_register_operator(const char *name, void *fn_init, void *fn_e
     }
 }
 
+/*
+* \brief Connetion hook to limit the number of
+* connections in BUSY state
+*
+* \param conn Pointer to connection struct
+*
+* \retval DECLINED On failure
+* \retval OK On Success
+*/
+static int hook_connection_early(conn_rec *conn)
+{
+    sb_handle *sb = conn->sbh;
+    int i, j;
+    unsigned long int ip_count = 0;
+    worker_score *ws_record = NULL;
+
+    if(sb != NULL && conn_read_state_limit > 0)   {
+
+        ws_record = &ap_scoreboard_image->servers[sb->child_num][sb->thread_num];
+        if(ws_record == NULL)
+            return DECLINED;
+
+        apr_cpystrn(ws_record->client, conn->remote_ip, sizeof(ws_record->client));
+
+        for (i = 0; i < server_limit; ++i) {
+            for (j = 0; j < thread_limit; ++j) {
+
+                ws_record = ap_get_scoreboard_worker(i, j);
+
+                if(ws_record == NULL)
+                    return DECLINED;
+
+                switch (ws_record->status) {
+                    case SERVER_BUSY_READ:
+                        if (strcmp(conn->remote_ip, ws_record->client) == 0)
+                            ip_count++;
+                        break;
+                    default:
+                        break;
+                }
+            }
+        }
+
+        if (ip_count > conn_read_state_limit) {
+            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL, "ModSecurity: Access denied with code 400. Too many threads [%ld] of %ld allowed in READ state from %s - Possible DoS Consumption Attack [Rejected]", ip_count,conn_read_state_limit,conn->remote_ip);
+            return OK;
+        } else {
+            return DECLINED;
+        }
+    }
+
+    return DECLINED;
+}
+
 /**
  * This function is exported for other Apache modules to
  * register new variables.
@@ -1051,7 +1135,7 @@ static void modsec_register_variable(const char *name, unsigned int type,
                                      void *fn_validate, void *fn_generate,
                                      unsigned int is_cacheable, unsigned int availability) {
     if (modsecurity != NULL) {
-        msre_engine_variable_register(modsecurity->msre, name, type, argc_min, argc_max, fn_validate, fn_generate, is_cacheable, availability);
+        msre_engine_variable_register(modsecurity->msre, name, type, argc_min, argc_max, (fn_var_validate_t)fn_validate, (fn_var_generate_t)fn_generate, is_cacheable, availability);
     }
     else {
         fprintf(stderr,"modsecurity is NULL\n");
@@ -1077,6 +1161,7 @@ static void register_hooks(apr_pool_t *mp) {
         "mod_rpaf.c",
         "mod_rpaf-2.0.c",
         "mod_extract_forwarded2.c",
+        "mod_remoteip.c",
         "mod_custom_header.c",
         "mod_breach_realip.c",
         "mod_breach_trans.c",
@@ -1098,6 +1183,10 @@ static void register_hooks(apr_pool_t *mp) {
     APR_REGISTER_OPTIONAL_FN(modsec_register_variable);
 #endif
 
+    /* For connection level hook */
+    ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
+    ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &server_limit);
+
     /* Main hooks */
     ap_hook_pre_config(hook_pre_config, NULL, NULL, APR_HOOK_FIRST);
     ap_hook_post_config(hook_post_config, postconfig_beforeme_list,
@@ -1108,6 +1197,9 @@ static void register_hooks(apr_pool_t *mp) {
      * // ap_hook_handler(hook_handler, NULL, NULL, APR_HOOK_MIDDLE);
      */
 
+    /* Connection processing hooks */
+    ap_hook_process_connection(hook_connection_early, NULL, NULL, APR_HOOK_FIRST);
+
     /* Transaction processing hooks */
     ap_hook_post_read_request(hook_request_early,
         postread_beforeme_list, postread_afterme_list, APR_HOOK_REALLY_FIRST);
@@ -1126,8 +1218,18 @@ static void register_hooks(apr_pool_t *mp) {
 
     ap_register_input_filter("MODSECURITY_IN", input_filter,
         NULL, AP_FTYPE_CONTENT_SET);
+
+    /* Ensure that the output filter runs before other modules so that
+     * we get a request that has a better chance of not being modified:
+     *
+     * Currently:
+     *   mod_expires = -2
+     *   mod_cache   = -1
+     *   mod_deflate = -1
+     *   mod_headers =  0
+     */
     ap_register_output_filter("MODSECURITY_OUT", output_filter,
-        NULL, AP_FTYPE_CONTENT_SET);
+        NULL, AP_FTYPE_CONTENT_SET - 3);
 
     ap_register_output_filter("PDFP_OUT", pdfp_output_filter,
         NULL, AP_FTYPE_CONTENT_SET);
