@@ -1,11 +1,19 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2008 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
  *
- * You should have received a copy of the licence along with this
- * program (stored in the file "LICENSE"). If the file is missing,
- * or if you have any other questions related to the licence, please
- * write to Breach Security, Inc. at support@breach.com.
+ * This product is released under the terms of the General Public Licence,
+ * version 2 (GPLv2). Please refer to the file LICENSE (included with this
+ * distribution) which contains the complete text of the licence.
+ *
+ * There are special exceptions to the terms and conditions of the GPL
+ * as it is applied to this software. View the full text of the exception in
+ * file MODSECURITY_LICENSING_EXCEPTION in the directory of this software
+ * distribution.
+ *
+ * If any of the files related to licensing are missing or if you have any
+ * other questions related to licensing please contact Trustwave Holdings, Inc.
+ * directly using the email address support@trustwave.com.
  *
  */
 #ifndef _MODSECURITY_H_
@@ -24,41 +32,34 @@ typedef struct msc_data_chunk msc_data_chunk;
 typedef struct msc_arg msc_arg;
 typedef struct msc_string msc_string;
 
-#if !(defined(WIN32) || defined(CYGWIN) || defined(NETWARE) || defined(SOLARIS2))
-#define DSOLOCAL __attribute__((visibility("hidden")))
-#else
-#define DSOLOCAL
-#endif
-
+#include "msc_release.h"
 #include "msc_logging.h"
 #include "msc_multipart.h"
 #include "msc_pcre.h"
 #include "msc_util.h"
-#ifdef WITH_LIBXML2
 #include "msc_xml.h"
-#endif
+#include "msc_geo.h"
 #include "re.h"
 
 #include "ap_config.h"
 #include "apr_md5.h"
 #include "apr_strings.h"
+#include "apr_hash.h"
 #include "httpd.h"
 #include "http_config.h"
 #include "http_log.h"
 #include "http_protocol.h"
 
-#define MODULE_NAME "ModSecurity"
-#define MODULE_RELEASE "2.1.7"
-#define MODULE_NAME_FULL (MODULE_NAME " v" MODULE_RELEASE " (Apache 2.x)")
-
 #define PHASE_REQUEST_HEADERS       1
 #define PHASE_REQUEST_BODY          2
 #define PHASE_RESPONSE_HEADERS      3
 #define PHASE_RESPONSE_BODY         4
 #define PHASE_LOGGING               5
+#define PHASE_FIRST                 PHASE_REQUEST_HEADERS
+#define PHASE_LAST                  PHASE_LOGGING
 
 #define NOT_SET                    -1l
-#define NOT_SET_P          (void *)-1l
+#define NOT_SET_P         ((void *)-1l)
 
 #define CREATEMODE ( APR_UREAD | APR_UWRITE | APR_GREAD )
 #define CREATEMODE_DIR ( APR_UREAD | APR_UWRITE | APR_UEXECUTE | APR_GREAD | APR_GEXECUTE )
@@ -82,9 +83,23 @@ typedef struct msc_string msc_string;
 #define REQUEST_BODY_HARD_LIMIT                 1073741824L
 #define REQUEST_BODY_DEFAULT_INMEMORY_LIMIT     131072
 #define REQUEST_BODY_DEFAULT_LIMIT              134217728
+#define REQUEST_BODY_NO_FILES_DEFAULT_LIMIT     1048576
 #define RESPONSE_BODY_DEFAULT_LIMIT             524288
 #define RESPONSE_BODY_HARD_LIMIT                1073741824L
 
+#define RESPONSE_BODY_LIMIT_ACTION_REJECT       0
+#define RESPONSE_BODY_LIMIT_ACTION_PARTIAL      1
+
+#define REQUEST_BODY_FORCEBUF_OFF               0
+#define REQUEST_BODY_FORCEBUF_ON                1
+
+#define SECACTION_TARGETS                       "REMOTE_ADDR"
+#define SECACTION_ARGS                          "@unconditionalMatch"
+
+#define SECMARKER_TARGETS                       "REMOTE_ADDR"
+#define SECMARKER_ARGS                          "@noMatch"
+#define SECMARKER_BASE_ACTIONS                  "t:none,pass,id:"
+
 #if !defined(OS2) && !defined(WIN32) && !defined(BEOS) && !defined(NETWARE)
 #include "unixd.h"
 #define __SET_MUTEX_PERMS
@@ -112,6 +127,12 @@ extern module AP_MODULE_DECLARE_DATA security2_module;
 
 extern DSOLOCAL const command_rec module_directives[];
 
+extern DSOLOCAL unsigned long int msc_pcre_match_limit;
+
+extern DSOLOCAL unsigned long int msc_pcre_match_limit_recursion;
+
+extern DSOLOCAL unsigned long int conn_read_state_limit;
+
 #define RESBODY_STATUS_NOT_READ         0   /* we were not configured to read the body */
 #define RESBODY_STATUS_ERROR            1   /* error occured while we were reading the body */
 #define RESBODY_STATUS_PARTIAL          2   /* partial body content available in the brigade */
@@ -136,11 +157,16 @@ extern DSOLOCAL const command_rec module_directives[];
 #define ACTION_PROXY                    3
 #define ACTION_DROP                     4
 #define ACTION_ALLOW                    5
+#define ACTION_ALLOW_REQUEST            6
+#define ACTION_ALLOW_PHASE              7
 
 #define MODSEC_DISABLED                 0
 #define MODSEC_DETECTION_ONLY           1
 #define MODSEC_ENABLED                  2
 
+#define MODSEC_CACHE_DISABLED           0
+#define MODSEC_CACHE_ENABLED            1
+
 #define MODSEC_OFFLINE                  0
 #define MODSEC_ONLINE                   1
 
@@ -186,12 +212,13 @@ struct modsec_rec {
     unsigned int         if_started_forwarding;
 
     apr_size_t           reqbody_length;
-    unsigned int         reqbody_status;
 
     apr_bucket_brigade  *of_brigade;
     unsigned int         of_status;
     unsigned int         of_done_reading;
     unsigned int         of_skipping;
+    unsigned int         of_partial;
+    unsigned int         of_is_error;
 
     unsigned int         resbody_status;
     apr_size_t           resbody_length;
@@ -225,7 +252,7 @@ struct modsec_rec {
     const char          *request_protocol;
 
     const char          *hostname;
-    
+
     apr_table_t         *request_headers;
 
     apr_off_t            request_content_length;
@@ -241,6 +268,9 @@ struct modsec_rec {
 
     apr_table_t         *tx_vars;
 
+    /* ENH: refactor to allow arbitrary var tables */
+    apr_table_t         *geo_vars;
+
     /* response */
     unsigned int         response_status;
     const char          *status_line;
@@ -262,7 +292,7 @@ struct modsec_rec {
     unsigned int         msc_reqbody_chunk_offset;   /* offset of the chunk currently in use     */
     msc_data_chunk      *msc_reqbody_chunk_current;  /* current chunk                            */
     char                *msc_reqbody_buffer;
-    
+
     const char          *msc_reqbody_filename;       /* when stored on disk */
     int                  msc_reqbody_fd;
     msc_data_chunk      *msc_reqbody_disk_chunk;
@@ -271,11 +301,11 @@ struct modsec_rec {
     int                  msc_reqbody_error;
     const char          *msc_reqbody_error_msg;
 
+    apr_size_t           msc_reqbody_no_files_length;
+
     multipart_data      *mpd;                        /* MULTIPART processor data structure */
 
-    #ifdef WITH_LIBXML2
     xml_data            *xml;                        /* XML processor data structure       */
-    #endif
 
     /* audit logging */
     char                *new_auditlog_boundary;
@@ -296,18 +326,43 @@ struct modsec_rec {
     apr_time_t           time_checkpoint_2;
     apr_time_t           time_checkpoint_3;
 
-    const char          *matched_var;
+    apr_array_header_t  *matched_rules;
+    msc_string          *matched_var;
+    int                  highest_severity;
 
     /* upload */
     int                  upload_extract_files;
     int                  upload_remove_files;
+    int                  upload_files_count;
 
     /* other */
+    apr_table_t         *collections_original;
     apr_table_t         *collections;
     apr_table_t         *collections_dirty;
 
     /* rule processing temp pool */
     apr_pool_t          *msc_rule_mptmp;
+
+    /* content injection */
+    const char          *content_prepend;
+    apr_off_t            content_prepend_len;
+    const char          *content_append;
+    apr_off_t            content_append_len;
+
+    /* data cache */
+    apr_hash_t          *tcache;
+    apr_size_t           tcache_items;
+
+    /* removed rules */
+    apr_array_header_t  *removed_rules;
+
+    /* When "allow" is executed the variable below is
+     * updated to contain the scope of the allow action. Set
+     * at 0 by default, it will have ACTION_ALLOW if we are
+     * to allow phases 1-4 and ACTION_ALLOW_REQUEST if we
+     * are to allow phases 1-2 only.
+     */
+    unsigned int         allow_scope;
 };
 
 struct directory_config {
@@ -317,13 +372,16 @@ struct directory_config {
 
     int                  is_enabled;
     int                  reqbody_access;
+    int                  reqbody_buffering;
     long int             reqbody_inmemory_limit;
     long int             reqbody_limit;
+    long int             reqbody_no_files_limit;
     int                  resbody_access;
 
     long int             of_limit;
     apr_table_t         *of_mime_types;
     int                  of_mime_types_cleared;
+    int                  of_limit_action;
 
     const char          *debuglog_name;
     int                  debuglog_level;
@@ -344,6 +402,10 @@ struct directory_config {
     /* AUDITLOG_SERIAL (single file) or AUDITLOG_CONCURRENT (multiple files) */
     int                  auditlog_type;
 
+    /* Mode for audit log directories and files */
+    apr_fileperms_t      auditlog_dirperms;
+    apr_fileperms_t      auditlog_fileperms;
+
     /* The name of the audit log file (for the old type), or the
      * name of the index file (for the new audit log type)
      */
@@ -370,22 +432,53 @@ struct directory_config {
     /* A regular expression that determines if a response
      * status is treated as relevant.
      */
-    msc_regex_t         *auditlog_relevant_regex;    
+    msc_regex_t         *auditlog_relevant_regex;
 
     /* Upload */
     const char          *tmp_dir;
     const char          *upload_dir;
     int                  upload_keep_files;
     int                  upload_validates_files;
-    int                  upload_filemode;
+    int                  upload_filemode; /* int only so NOT_SET works */
+    int                  upload_file_limit;
 
     /* Used only in the configuration phase. */
     msre_rule           *tmp_chain_starter;
     msre_actionset      *tmp_default_actionset;
+    apr_table_t         *tmp_rule_placeholders;
 
     /* Misc */
     const char          *data_dir;
     const char          *webappid;
+
+    /* Content injection. */
+    int                  content_injection_enabled;
+
+    /* PDF XSS Protection. */
+    int                  pdfp_enabled;
+    const char          *pdfp_secret;
+    int                  pdfp_timeout;
+    const char          *pdfp_token_name;
+    int                  pdfp_only_get;
+    int                  pdfp_method;
+
+    /* Geo Lookup */
+    geo_db              *geo;
+
+    /* Cache */
+    int                  cache_trans;
+    int                  cache_trans_incremental;
+    apr_size_t           cache_trans_min;
+    apr_size_t           cache_trans_max;
+    apr_size_t           cache_trans_maxitems;
+
+    /* Array to hold signatures of components, which will
+     * appear in the ModSecurity signature in the audit log.
+     */
+    apr_array_header_t  *component_signatures;
+
+    /* Request character encoding. */
+    const char          *request_encoding;
 };
 
 struct error_message {
@@ -399,6 +492,7 @@ struct error_message {
 struct msc_engine {
     apr_pool_t              *mp;
     apr_global_mutex_t      *auditlog_lock;
+    apr_global_mutex_t      *geo_lock;
     msre_engine             *msre;
     unsigned int             processing_mode;
 };
@@ -441,19 +535,19 @@ void DSOLOCAL modsecurity_shutdown(msc_engine *msce);
 
 apr_status_t DSOLOCAL modsecurity_tx_init(modsec_rec *msr);
 
-apr_status_t DSOLOCAL modsecurity_process_phase(modsec_rec *msr, int phase);
+apr_status_t DSOLOCAL modsecurity_process_phase(modsec_rec *msr, unsigned int phase);
 
 
 /* Request body functions */
 
-apr_status_t DSOLOCAL modsecurity_request_body_start(modsec_rec *msr);
+apr_status_t DSOLOCAL modsecurity_request_body_start(modsec_rec *msr, char **error_msg);
 
 apr_status_t DSOLOCAL modsecurity_request_body_store(modsec_rec *msr,
-    const char *data, apr_size_t length);
+    const char *data, apr_size_t length, char **error_msg);
 
-apr_status_t DSOLOCAL modsecurity_request_body_end(modsec_rec *msr);
+apr_status_t DSOLOCAL modsecurity_request_body_end(modsec_rec *msr, char **error_msg);
 
-apr_status_t DSOLOCAL modsecurity_request_body_retrieve_start(modsec_rec *msr);
+apr_status_t DSOLOCAL modsecurity_request_body_retrieve_start(modsec_rec *msr, char **error_msg);
 
 apr_status_t DSOLOCAL modsecurity_request_body_retrieve_end(modsec_rec *msr);
 
@@ -462,14 +556,17 @@ apr_status_t DSOLOCAL modsecurity_request_body_retrieve_end(modsec_rec *msr);
  * nbytes will contain the number of bytes stored in the buffer.
  */
 apr_status_t DSOLOCAL modsecurity_request_body_retrieve(modsec_rec *msr, msc_data_chunk **chunk,
-    long int nbytes);
+    long int nbytes, char **error_msg);
 
 void DSOLOCAL msc_add(modsec_rec *msr, int level, msre_actionset *actionset,
     const char *action_message, const char *rule_message);
 
+const char DSOLOCAL *msc_alert_message(modsec_rec *msr, msre_actionset *actionset, const char *action_message,
+    const char *rule_message);
+
 void DSOLOCAL msc_alert(modsec_rec *msr, int level, msre_actionset *actionset, const char *action_message,
     const char *rule_message);
 
-apr_status_t DSOLOCAL modsecurity_request_body_clear(modsec_rec *msr);
+apr_status_t DSOLOCAL modsecurity_request_body_clear(modsec_rec *msr, char **error_msg);
 
 #endif
