@@ -1,6 +1,6 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2010 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
  *
  * This product is released under the terms of the General Public Licence,
  * version 2 (GPLv2). Please refer to the file LICENSE (included with this
@@ -12,8 +12,8 @@
  * distribution.
  *
  * If any of the files related to licensing are missing or if you have any
- * other questions related to licensing please contact Breach Security, Inc.
- * directly using the email address support@breach.com.
+ * other questions related to licensing please contact Trustwave Holdings, Inc.
+ * directly using the email address support@trustwave.com.
  *
  */
 #if defined(WITH_LUA)
@@ -100,7 +100,7 @@ char *lua_compile(msc_script **script, const char *filename, apr_pool_t *pool) {
     /* Find script. */
     if (luaL_loadfile(L, filename)) {
         return apr_psprintf(pool, "ModSecurity: Failed to compile script %s: %s",
-            filename, lua_tostring(L, -1));
+                filename, lua_tostring(L, -1));
     }
 
     /* Dump the script into binary form. */
@@ -234,7 +234,7 @@ static int l_getvar(lua_State *L) {
 
     /* Resolve variable. */
     var = msre_create_var_ex(msr->msc_rule_mptmp, msr->modsecurity->msre,
-        varname, param, msr, &my_error_msg);
+            varname, param, msr, &my_error_msg);
 
     if (var == NULL) {
         msr_log(msr, 1, "%s", my_error_msg);
@@ -303,7 +303,7 @@ static int l_getvars(lua_State *L) {
 
     /* Resolve variable. */
     vartemplate = msre_create_var_ex(msr->msc_rule_mptmp, msr->modsecurity->msre,
-        varname, param, msr, &my_error_msg);
+            varname, param, msr, &my_error_msg);
 
     if (vartemplate == NULL) {
         msr_log(msr, 1, "%s", my_error_msg);
@@ -337,10 +337,58 @@ static int l_getvars(lua_State *L) {
     return 1;
 }
 
+/*
+* \brief New setvar function for Lua API. Users can put back
+* data in modsecurity core via new variables
+*
+* \param L Pointer to Lua state
+*
+* \retval -1 On failure
+* \retval 0 On Collection failure
+* \retval 1 On Success
+*/
+static int l_setvar(lua_State *L) {
+    modsec_rec *msr = NULL;
+    msre_rule *rule = NULL;
+    const char *var_value = NULL;
+    const char *var_name = NULL;
+    int nargs = lua_gettop(L);
+    char *chr = NULL;
+
+    lua_getglobal(L, "__msr");
+    msr = (modsec_rec *)lua_topointer(L, -1);
+
+    lua_getglobal(L, "__rule");
+    rule = (msre_rule *)lua_topointer(L, -1);
+
+    if(nargs != 2)  {
+        msr_log(msr, 8, "m.setvar: Failed m.setvar funtion must has 2 arguments");
+        return -1;
+    }
+
+    var_value = luaL_checkstring (L, 2);
+    var_name = luaL_checkstring (L, 1);
+
+    lua_pop(L,2);
+
+    if(var_value == NULL || var_name == NULL)
+        return -1;
+
+    chr = strchr((char *)var_name,0x2e);
+
+    if(chr == NULL) {
+        msr_log(msr, 8, "m.setvar: Must specify a collection using dot character - ie m.setvar(tx.myvar,mydata)");
+        return -1;
+    }
+
+    return msre_action_setvar_execute(msr,msr->msc_rule_mptmp,rule,(char *)var_name,(char *)var_value);
+}
+
 static const struct luaL_Reg mylib[] = {
     { "log", l_log },
     { "getvar", l_getvar },
     { "getvars", l_getvars },
+    { "setvar", l_setvar },
     { NULL, NULL }
 };
 
@@ -376,15 +424,22 @@ int lua_execute(msc_script *script, char *param, modsec_rec *msr, msre_rule *rul
         lua_setglobal(L, "__rule");
     }
 
+    //lua_pushcfunction(L, l_setvar);
+    //lua_setglobal(L, "mysetvar");
+
     /* Register functions. */
     luaL_register(L, "m", mylib);
 
     rc = lua_restore(L, script);
     if (rc) {
         *error_msg = apr_psprintf(msr->mp, "Lua: Failed to restore script with %i.", rc);
+
+        if (msr->txcfg->debuglog_level >= 8) {
+            msr_log(msr, 8, "Lua: Failed to restore script with: %d",rc);
+        }
+
         return -1;
     }
-
     /* Execute the chunk so that the functions are defined. */
     lua_pcall(L, 0, 0, 0);
 
@@ -397,7 +452,13 @@ int lua_execute(msc_script *script, char *param, modsec_rec *msr, msre_rule *rul
     }
 
     if (lua_pcall(L, ((param != NULL) ? 1 : 0), 1, 0)) {
+
         *error_msg = apr_psprintf(msr->mp, "Lua: Script execution failed: %s", lua_tostring(L, -1));
+
+        if (msr->txcfg->debuglog_level >= 8) {
+            msr_log(msr, 8, "Lua: Script execution failed: %s", lua_tostring(L, -1));
+        }
+
         return -1;
     }
 
@@ -414,7 +475,7 @@ int lua_execute(msc_script *script, char *param, modsec_rec *msr, msre_rule *rul
     /* Returns status code to caller. */
     if (msr->txcfg->debuglog_level >= 8) {
         msr_log(msr, 8, "Lua: Script completed in %" APR_TIME_T_FMT " usec, returning: %s.",
-            (apr_time_now() - time_before), *error_msg);
+                (apr_time_now() - time_before), *error_msg);
     }
 
     return ((*error_msg != NULL) ? RULE_MATCH : RULE_NO_MATCH);
