@@ -1,6 +1,6 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2009 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
  *
  * This product is released under the terms of the General Public Licence,
  * version 2 (GPLv2). Please refer to the file LICENSE (included with this
@@ -12,27 +12,67 @@
  * distribution.
  *
  * If any of the files related to licensing are missing or if you have any
- * other questions related to licensing please contact Breach Security, Inc.
- * directly using the email address support@breach.com.
+ * other questions related to licensing please contact Trustwave Holdings, Inc.
+ * directly using the email address support@trustwave.com.
  *
  */
-#include "msc_release.h"
-#include "msc_util.h"
-
 #include <ctype.h>
 #include <fcntl.h>
 #include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#include "mod_security2_config.h"
+#include "msc_release.h"
+#include "msc_util.h"
+
 #include <apr_lib.h>
 
-/* NOTE: Be careful as these can ONLY be used on static values for X.
+/**
+ * NOTE: Be careful as these can ONLY be used on static values for X.
  * (i.e. VALID_HEX(c++) will NOT work)
  */
 #define VALID_HEX(X) (((X >= '0')&&(X <= '9')) || ((X >= 'a')&&(X <= 'f')) || ((X >= 'A')&&(X <= 'F')))
 #define ISODIGIT(X) ((X >= '0')&&(X <= '7'))
 
+#if (defined(WIN32) || defined(NETWARE))
+/** Windows does not define all the octal modes */
+#define S_IXOTH 00001
+#define S_IWOTH 00002
+#define S_IROTH 00004
+#define S_IXGRP 00010
+#define S_IWGRP 00020
+#define S_IRGRP 00040
+#define S_IXUSR 00100
+#define S_IWUSR 00200
+#define S_IRUSR 00400
+#define S_ISVTX 01000
+#define S_ISGID 02000
+#define S_ISUID 04000
+#endif /* defined(WIN32 || NETWARE) */
+
+/* Base64 tables used in decodeBase64Ext */
+static const char b64_pad = '=';
+
+static const short b64_reverse_t[256] = {
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -2, -2, -1, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, 62, -2, -2, -2, 63,
+  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -2, -2, -2, -2, -2, -2,
+  -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
+  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -2, -2, -2, -2, -2,
+  -2, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
+  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
+  -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2
+};
+
 
 /**
  *
@@ -49,6 +89,72 @@ int parse_boolean(const char *input) {
     return -1;
 }
 
+/* \brief Decode Base64 data with special chars
+*
+* \param plain_text Pointer to plain text data
+* \param input Pointer to input data
+* \param input_len Input data length
+*
+* \retval 0 On failure
+* \retval string length On Success
+*/
+int decode_base64_ext(char *plain_text, const char *input, int input_len)
+{
+    const char *encoded = input;
+    int i = 0, j = 0, k = 0;
+    int ch = 0;
+
+    while ((ch = *encoded++) != '\0' && input_len-- > 0) {
+        if (ch == b64_pad) {
+            if (*encoded != '=' && (i % 4) == 1) {
+                return 0;
+            }
+            continue;
+        }
+
+        ch = b64_reverse_t[ch];
+        if (ch < 0 || ch == -1) {
+            continue;
+        } else if (ch == -2) {
+            return 0;
+        }
+        switch(i % 4) {
+            case 0:
+                plain_text[j] = ch << 2;
+                break;
+            case 1:
+                plain_text[j++] |= ch >> 4;
+                plain_text[j] = (ch & 0x0f) << 4;
+                break;
+            case 2:
+                plain_text[j++] |= ch >>2;
+                plain_text[j] = (ch & 0x03) << 6;
+                break;
+            case 3:
+                plain_text[j++] |= ch;
+                break;
+        }
+        i++;
+    }
+
+    k = j;
+    if (ch == b64_pad) {
+        switch(i % 4) {
+            case 1:
+                return 0;
+            case 2:
+                k++;
+            case 3:
+                plain_text[k] = 0;
+        }
+    }
+
+    plain_text[j] = '\0';
+
+    return j;
+}
+
+
 /**
  * Parses a string that contains a name-value pair in the form "name=value".
  * IMP1 It does not check for whitespace between tokens.
@@ -418,15 +524,33 @@ char *current_filetime(apr_pool_t *mp) {
 /**
  *
  */
-int msc_mkstemp(char *template) {
+int msc_mkstemp_ex(char *template, int mode) {
+    int fd = -1;
+
     /* ENH Use apr_file_mktemp instead. */
 
-    #if !(defined(WIN32)||defined(NETWARE))
-    return mkstemp(template);
-    #else
+#if !(defined(WIN32)||defined(NETWARE))
+    fd = mkstemp(template);
+#ifdef HAVE_FCHMOD
+    if ((fd != -1) && (mode != 0)) {
+        if (fchmod(fd, mode) == -1) {
+            return -1;
+        }
+    }
+#endif /* HAVE_FCHMOD */
+#else
     if (mktemp(template) == NULL) return -1;
-    return open(template, O_WRONLY | O_APPEND | O_CREAT | O_BINARY, CREATEMODE_UNISTD);
-    #endif
+    fd = open(template, O_WRONLY | O_APPEND | O_CREAT | O_BINARY, mode);
+#endif /* !(defined(WIN32)||defined(NETWARE)) */
+
+    return fd;
+}
+
+/**
+ *
+ */
+int msc_mkstemp(char *template) {
+    return msc_mkstemp_ex(template, CREATEMODE_UNISTD);
 }
 
 /**
@@ -493,6 +617,27 @@ char *log_escape_raw(apr_pool_t *mp, const unsigned char *text, unsigned long in
     return (char *)ret;
 }
 
+char *log_escape_nul(apr_pool_t *mp, const unsigned char *text, unsigned long int text_length) {
+    unsigned char *ret = apr_palloc(mp, text_length * 4 + 1);
+    unsigned long int i, j;
+
+    for (i = 0, j = 0; i < text_length; i++) {
+        if (text[i] == '\0') {
+            ret[j] = '\\';
+            ret[j+1] = 'x';
+            c2x(text[i], ret+j+2);
+            j += 4;
+        }
+        else {
+            ret[j] = text[i];
+            j++;
+        }
+    }
+    ret[j] = '\0';
+
+    return (char *)ret;
+}
+
 /**
  * Transform text to ASCII printable or hex escaped
  */
@@ -662,7 +807,7 @@ int js_decode_nonstrict_inplace(unsigned char *input, long int input_len) {
                         j = 2;
                         buf[j] = '\0';
                     }
-                    *d++ = strtol(buf, NULL, 8);
+                    *d++ = (unsigned char)strtol(buf, NULL, 8);
                     i += 1 + j;
                     count++;
                 }
@@ -1105,68 +1250,175 @@ int ansi_c_sequences_decode_inplace(unsigned char *input, int input_len) {
  * IMP1 Assumes NUL-terminated
  */
 int normalise_path_inplace(unsigned char *input, int input_len, int win, int *changed) {
-    unsigned char *d = input;
-    int i, count;
+    unsigned char *src;
+    unsigned char *dst;
+    unsigned char *end;
+    int ldst = 0;
+    int hitroot = 0;
+    int done = 0;
+    int relative;
+    int trailing;
 
     *changed = 0;
 
-    i = count = 0;
-    while ((i < input_len)&&(count < input_len)) {
-        char c = input[i];
+    /* Need at least one byte to normalize */
+    if (input_len <= 0) return 0;
 
+    /*
+     * ENH: Deal with UNC and drive letters?
+     */
+
+    src = dst = input;
+    end = input + (input_len - 1);
+    ldst = 1;
+
+    relative = ((*input == '/') || (win && (*input == '\\'))) ? 0 : 1;
+    trailing = ((*end == '/') || (win && (*end == '\\'))) ? 1 : 0;
+
+
+    while (!done && (src <= end) && (dst <= end)) {
         /* Convert backslash to forward slash on Windows only. */
-        if ((win)&&(c == '\\')) {
-            c = '/';
-            *changed = 1;
-        }
-
-        if (c == '/') {
-            /* Is there a directory back-reference? Yes, we
-             * require at least 5 prior bytes here. That's on
-             * purpose.
-             */
-            if ((count >= 5)&&(*(d - 1) == '.')&&(*(d - 2) == '.')&&(*(d - 3) == '/')) {
-                unsigned char *cd = d - 4;
-                int ccount = count - 4;
-
+        if (win) {
+            if (*src == '\\') {
+                *src = '/';
                 *changed = 1;
-
-                /* Go back until we reach the beginning or a forward slash. */
-                while ((ccount > 0)&&(*cd != '/')) {
-                    ccount--;
-                    cd--;
-                }
-
-                if (*cd == '/') {
-                    d = cd;
-                    count = ccount;
-                }
-            } else
-            /* Is there a directory self-reference? */
-            if ((count >= 2)&&(*(d - 1) == '.')&&(*(d - 2) == '/')) {
-                /* Ignore the last two bytes. */
-                d -= 2;
-                count -= 2;
-                *changed = 1;
-            } else
-            /* Or are there just multiple occurences of forward slash? */
-            if ((count >= 1)&&(*(d - 1) == '/')) {
-                /* Ignore the last one byte. */
-                d--;
-                count--;
+            }
+            if ((src < end) && (*(src + 1) == '\\')) {
+                *(src + 1) = '/';
                 *changed = 1;
             }
         }
 
-        /* Copy the byte over. */
-        *d++ = c;
-        count++;
-        i++;
+        /* Always normalize at the end of the input. */
+        if (src == end) {
+            done = 1;
+        }
+
+        /* Skip normalization if this is NOT the end of the path segment. */
+        else if (*(src + 1) != '/') {
+            goto copy; /* Skip normalization. */
+        }
+
+        /*** Normalize the path segment. ***/
+
+        /* Could it be an empty path segment? */
+        if ((src != end) && *src == '/') {
+            /* Ignore */
+            *changed = 1;
+            goto copy; /* Copy will take care of this. */
+        }
+
+        /* Could it be a back or self reference? */
+        else if (*src == '.') {
+
+            /* Back-reference? */
+            if ((dst > input) && (*(dst - 1) == '.')) {
+                /* If a relative path and either our normalization has
+                 * already hit the rootdir, or this is a backref with no
+                 * previous path segment, then mark that the rootdir was hit
+                 * and just copy the backref as no normilization is possible.
+                 */
+                if (relative && (hitroot || ((dst - 2) <= input))) {
+                    hitroot = 1;
+
+                    goto copy; /* Skip normalization. */
+                }
+
+                /* Remove backreference and the previous path segment. */
+                dst -= 3;
+                while ((dst > input) && (*dst != '/')) {
+                    dst--;
+                }
+
+                /* But do not allow going above rootdir. */
+                if (dst <= input) {
+                    hitroot = 1;
+                    dst = input;
+
+                    /* Need to leave the root slash if this
+                     * is not a relative path and the end was reached
+                     * on a backreference.
+                     */
+                    if (!relative && (src == end)) {
+                        dst++;
+                    }
+                }
+
+                if (done) goto length; /* Skip the copy. */
+                src++;
+
+                *changed = 1;
+            }
+
+            /* Relative Self-reference? */
+            else if (dst == input) {
+                *changed = 1;
+
+                /* Ignore. */
+
+                if (done) goto length; /* Skip the copy. */
+                src++;
+            }
+
+            /* Self-reference? */
+            else if (*(dst - 1) == '/') {
+                *changed = 1;
+
+                /* Ignore. */
+
+                if (done) goto length; /* Skip the copy. */
+                dst--;
+                src++;
+            }
+        }
+
+        /* Found a regular path segment. */
+        else if (dst > input) {
+            hitroot = 0;
+        }
+
+copy:
+        /*** Copy the byte if required. ***/
+
+        /* Skip to the last forward slash when multiple are used. */
+        if (*src == '/') {
+            unsigned char *oldsrc = src;
+
+            while (   (src < end)
+                   && ((*(src + 1) == '/') || (win && (*(src + 1) == '\\'))) )
+            {
+                src++;
+            }
+            if (oldsrc != src) *changed = 1;
+
+            /* Do not copy the forward slash to the root
+             * if it is not a relative path.  Instead
+             * move over the slash to the next segment.
+             */
+            if (relative && (dst == input)) {
+                src++;
+                goto length; /* Skip the copy */
+            }
+        }
+
+        *(dst++) = *(src++);
+
+length:
+        ldst = (dst - input);
     }
 
-    *d = '\0';
+    /* Make sure that there is not a trailing slash in the
+     * normalized form if there was not one in the original form.
+     */
+    if (!trailing && (dst > input) && *(dst - 1) == '/') {
+        ldst--;
+        dst--;
+    }
 
-    return count;
+    /* Always NUL terminate */
+    *dst = '\0';
+
+    return ldst;
 }
 
 char *modsec_build(apr_pool_t *mp) {
@@ -1351,3 +1603,26 @@ int css_decode_inplace(unsigned char *input, long int input_len) {
 
     return count;
 }
+
+/**
+ * Translate UNIX octal umask/mode to APR apr_fileperms_t
+ */
+apr_fileperms_t mode2fileperms(int mode) {
+    apr_fileperms_t perms = 0;
+
+    if (mode & S_IXOTH) perms |= APR_WEXECUTE;
+    if (mode & S_IWOTH) perms |= APR_WWRITE;
+    if (mode & S_IROTH) perms |= APR_WREAD;
+    if (mode & S_IXGRP) perms |= APR_GEXECUTE;
+    if (mode & S_IWGRP) perms |= APR_GWRITE;
+    if (mode & S_IRGRP) perms |= APR_GREAD;
+    if (mode & S_IXUSR) perms |= APR_UEXECUTE;
+    if (mode & S_IWUSR) perms |= APR_UWRITE;
+    if (mode & S_IRUSR) perms |= APR_UREAD;
+    if (mode & S_ISVTX) perms |= APR_WSTICKY;
+    if (mode & S_ISGID) perms |= APR_GSETID;
+    if (mode & S_ISUID) perms |= APR_USETID;
+
+    return perms;
+}
+
