@@ -1,11 +1,19 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2008 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
  *
- * You should have received a copy of the licence along with this
- * program (stored in the file "LICENSE"). If the file is missing,
- * or if you have any other questions related to the licence, please
- * write to Breach Security, Inc. at support@breach.com.
+ * This product is released under the terms of the General Public Licence,
+ * version 2 (GPLv2). Please refer to the file LICENSE (included with this
+ * distribution) which contains the complete text of the licence.
+ *
+ * There are special exceptions to the terms and conditions of the GPL
+ * as it is applied to this software. View the full text of the exception in
+ * file MODSECURITY_LICENSING_EXCEPTION in the directory of this software
+ * distribution.
+ *
+ * If any of the files related to licensing are missing or if you have any
+ * other questions related to licensing please contact Trustwave Holdings, Inc.
+ * directly using the email address support@trustwave.com.
  *
  */
 #include "persist_dbm.h"
@@ -14,7 +22,7 @@
 /**
  *
  */
-static apr_table_t *collection_unpack(modsec_rec *msr, char *blob, unsigned int blob_size,
+static apr_table_t *collection_unpack(modsec_rec *msr, const unsigned char *blob, unsigned int blob_size,
     int log_vars)
 {
     apr_table_t *col = NULL;
@@ -30,11 +38,33 @@ static apr_table_t *collection_unpack(modsec_rec *msr, char *blob, unsigned int
         msc_string *var = apr_pcalloc(msr->mp, sizeof(msc_string));
 
         var->name_len = (blob[blob_offset] << 8) + blob[blob_offset + 1];
-        if (var->name_len == 0) break;
+        if (var->name_len == 0) {
+            /* Is the length a name length, or just the end of the blob? */
+            if (blob_offset < blob_size - 2) {
+                /* This should never happen as the name length
+                 * includes the terminating NUL and should be 1 for ""
+                 */
+                if (msr->txcfg->debuglog_level >= 9) {
+                    msr_log(msr, 9, "BLOB[%d]: %s", blob_offset, log_escape_hex(msr->mp, blob + blob_offset, blob_size - blob_offset));
+                }
+                msr_log(msr, 4, "Possibly corrupted database: var name length = 0 at blob offset %u-%u.", blob_offset, blob_offset + 1);
+            }
+            break;
+        }
+        else if (var->name_len > 65536) {
+            /* This should never happen as the length is restricted on store
+             * to 65536.
+             */
+            if (msr->txcfg->debuglog_level >= 9) {
+                msr_log(msr, 9, "BLOB[%d]: %s", blob_offset, log_escape_hex(msr->mp, blob + blob_offset, blob_size - blob_offset));
+            }
+            msr_log(msr, 4, "Possibly corrupted database: var name length > 65536 (0x%04x) at blob offset %u-%u.", var->name_len, blob_offset, blob_offset + 1);
+            break;
+        }
 
         blob_offset += 2;
         if (blob_offset + var->name_len > blob_size) return NULL;
-        var->name = apr_pstrmemdup(msr->mp, blob + blob_offset, var->name_len - 1);
+        var->name = apr_pstrmemdup(msr->mp, (const char *)blob + blob_offset, var->name_len - 1);
         blob_offset += var->name_len;
         var->name_len--;
 
@@ -42,11 +72,11 @@ static apr_table_t *collection_unpack(modsec_rec *msr, char *blob, unsigned int
         blob_offset += 2;
 
         if (blob_offset + var->value_len > blob_size) return NULL;
-        var->value = apr_pstrmemdup(msr->mp, blob + blob_offset, var->value_len - 1);
+        var->value = apr_pstrmemdup(msr->mp, (const char *)blob + blob_offset, var->value_len - 1);
         blob_offset += var->value_len;
         var->value_len--;
 
-        if (log_vars) {
+        if (log_vars && (msr->txcfg->debuglog_level >= 9)) {
             msr_log(msr, 9, "Read variable: name \"%s\", value \"%s\".",
                 log_escape_ex(msr->mp, var->name, var->name_len),
                 log_escape_ex(msr->mp, var->value, var->value_len));
@@ -61,54 +91,72 @@ static apr_table_t *collection_unpack(modsec_rec *msr, char *blob, unsigned int
 /**
  *
  */
-apr_table_t *collection_retrieve(modsec_rec *msr, const char *col_name,
+static apr_table_t *collection_retrieve_ex(apr_sdbm_t *existing_dbm, modsec_rec *msr, const char *col_name,
     const char *col_key, int col_key_len)
 {
     char *dbm_filename = NULL;
     apr_status_t rc;
     apr_sdbm_datum_t key;
     apr_sdbm_datum_t *value = NULL;
-    apr_sdbm_t *dbm;
+    apr_sdbm_t *dbm = NULL;
     apr_table_t *col = NULL;
     const apr_array_header_t *arr;
     apr_table_entry_t *te;
+    int expired = 0;
     int i;
 
     if (msr->txcfg->data_dir == NULL) {
         msr_log(msr, 1, "Unable to retrieve collection (name \"%s\", key \"%s\"). Use "
             "SecDataDir to define data directory first.", log_escape(msr->mp, col_name),
-            log_escape(msr->mp, col_key));
-        return NULL;
+            log_escape_ex(msr->mp, col_key, col_key_len));
+        goto cleanup;
     }
 
     dbm_filename = apr_pstrcat(msr->mp, msr->txcfg->data_dir, "/", col_name, NULL);
 
-    rc = apr_sdbm_open(&dbm, dbm_filename, APR_READ | APR_SHARELOCK,
-        CREATEMODE, msr->mp);
-    if (rc != APR_SUCCESS) {
-        return NULL;
-    }
-
     key.dptr = (char *)col_key;
     key.dsize = col_key_len + 1;
 
+    if (existing_dbm == NULL) {
+        rc = apr_sdbm_open(&dbm, dbm_filename, APR_READ | APR_SHARELOCK,
+            CREATEMODE, msr->mp);
+        if (rc != APR_SUCCESS) {
+            dbm = NULL;
+            goto cleanup;
+        }
+    }
+    else {
+        dbm = existing_dbm;
+    }
+
     value = (apr_sdbm_datum_t *)apr_pcalloc(msr->mp, sizeof(apr_sdbm_datum_t));
     rc = apr_sdbm_fetch(dbm, value, key);
     if (rc != APR_SUCCESS) {
-        apr_sdbm_close(dbm);
         msr_log(msr, 1, "Failed to read from DBM file \"%s\": %s", log_escape(msr->mp,
             dbm_filename), get_apr_error(msr->mp, rc));
-        return NULL;
-    }
-    
-    if (value->dptr == NULL) { /* Key not found in DBM file. */
-        apr_sdbm_close(dbm);
-        return NULL;
+        goto cleanup;
     }
 
+    if (value->dptr == NULL) { /* Key not found in DBM file. */
+        goto cleanup;
+    }
+
+    /* ENH Need expiration (and perhaps other metadata) accessible in blob
+     * form to determine if converting to a table is needed.  This will
+     * save some cycles.
+     */
+
     /* Transform raw data into a table. */
-    col = collection_unpack(msr, value->dptr, value->dsize, 1);
-    if (col == NULL) return NULL;
+    col = collection_unpack(msr, (const unsigned char *)value->dptr, value->dsize, 1);
+    if (col == NULL) {
+        goto cleanup;
+    }
+
+    /* Close after "value" used from fetch or memory may be overwritten. */
+    if (existing_dbm == NULL) {
+        apr_sdbm_close(dbm);
+        dbm = NULL;
+    }
 
     /* Remove expired variables. */
     do {
@@ -119,22 +167,72 @@ apr_table_t *collection_retrieve(modsec_rec *msr, const char *col_name,
                 msc_string *var = (msc_string *)te[i].val;
                 int expiry_time = atoi(var->value);
 
-                /* Do not remove the record itself. */
-                if (strcmp(te[i].key, "__expire_KEY") == 0) continue;                
-
                 if (expiry_time <= apr_time_sec(msr->request_time)) {
-                    char *key_to_expire = apr_pstrdup(msr->mp, te[i].key);
-                    msr_log(msr, 9, "Removing key \"%s\" from collection.", key_to_expire + 9);
+                    char *key_to_expire = te[i].key;
+
+                    /* Done early if the col expired */
+                    if (strcmp(key_to_expire, "__expire_KEY") == 0) {
+                        expired = 1;
+                    }
+                    if (msr->txcfg->debuglog_level >= 9) {
+                        msr_log(msr, 9, "Removing key \"%s\" from collection.", key_to_expire + 9);
+                        msr_log(msr, 9, "Removing key \"%s\" from collection.", key_to_expire);
+                    }
                     apr_table_unset(col, key_to_expire + 9);
-                    msr_log(msr, 9, "Removing key \"%s\" from collection.", key_to_expire);
                     apr_table_unset(col, key_to_expire);
-                    msr_log(msr, 4, "Removed expired variable \"%s\".", key_to_expire + 9);
+                    if (msr->txcfg->debuglog_level >= 4) {
+                        msr_log(msr, 4, "Removed expired variable \"%s\".", key_to_expire + 9);
+                    }
                     break;
                 }
             }
         }
-    } while(i != arr->nelts);
-    
+    } while(!expired && (i != arr->nelts));
+
+    /* Delete the collection if the variable "KEY" does not exist.
+     *
+     * ENH It would probably be more efficient to hold the DBM
+     * open until determined if it needs deleted than to open a second
+     * time.
+     */
+    if (apr_table_get(col, "KEY") == NULL) {
+        if (existing_dbm == NULL) {
+            rc = apr_sdbm_open(&dbm, dbm_filename, APR_CREATE | APR_WRITE | APR_SHARELOCK,
+                CREATEMODE, msr->mp);
+            if (rc != APR_SUCCESS) {
+                msr_log(msr, 1, "Failed to access DBM file \"%s\": %s",
+                    log_escape(msr->mp, dbm_filename), get_apr_error(msr->mp, rc));
+                dbm = NULL;
+                goto cleanup;
+            }
+        }
+        else {
+            dbm = existing_dbm;
+        }
+
+        rc = apr_sdbm_delete(dbm, key);
+        if (rc != APR_SUCCESS) {
+            msr_log(msr, 1, "Failed deleting collection (name \"%s\", "
+                "key \"%s\"): %s", log_escape(msr->mp, col_name),
+                log_escape_ex(msr->mp, col_key, col_key_len), get_apr_error(msr->mp, rc));
+            goto cleanup;
+        }
+
+
+        if (existing_dbm == NULL) {
+            apr_sdbm_close(dbm);
+            dbm = NULL;
+        }
+
+        if (expired && (msr->txcfg->debuglog_level >= 9)) {
+            msr_log(msr, 9, "Collection expired (name \"%s\", key \"%s\").", col_name, log_escape_ex(msr->mp, col_key, col_key_len));
+        }
+        if (msr->txcfg->debuglog_level >= 4) {
+            msr_log(msr, 4, "Deleted collection (name \"%s\", key \"%s\").",
+                log_escape(msr->mp, col_name), log_escape_ex(msr->mp, col_key, col_key_len));
+        }
+        goto cleanup;
+    }
 
     /* Update UPDATE_RATE */
     {
@@ -150,38 +248,62 @@ apr_table_t *collection_retrieve(modsec_rec *msr, const char *col_name,
             if (var == NULL) {
                 /* Error. */
             } else {
-                int td;
+                apr_time_t td;
                 counter = atoi(var->value);
-                var = (msc_string *)apr_table_get(col, "UPDATE_RATE");
-                if (var == NULL) {
-                    var = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
-                    var->name = "UPDATE_RATE";
-                    var->name_len = strlen(var->name);
-                    apr_table_setn(col, var->name, (void *)var);
-                }
+
+                /* UPDATE_RATE is removed on store, so add it back here */
+                var = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
+                var->name = "UPDATE_RATE";
+                var->name_len = strlen(var->name);
+                apr_table_setn(col, var->name, (void *)var);
 
                 /* NOTE: No rate if there has been no time elapsed */
-                td = (int)(apr_time_sec(apr_time_now()) - create_time);
+                td = (apr_time_sec(apr_time_now()) - create_time);
                 if (td == 0) {
-                    var->value = apr_psprintf(msr->mp, "%i", 0);
+                    var->value = apr_psprintf(msr->mp, "%d", 0);
                 }
                 else {
-                    var->value = apr_psprintf(msr->mp, "%i",
-                        (int)((60 * counter)/td));
+                    var->value = apr_psprintf(msr->mp, "%" APR_TIME_T_FMT,
+                        (apr_time_t)((60 * counter)/td));
                 }
                 var->value_len = strlen(var->value);
             }
         }
     }
 
-    apr_sdbm_close(dbm);
+    if (msr->txcfg->debuglog_level >= 4) {
+        msr_log(msr, 4, "Retrieved collection (name \"%s\", key \"%s\").",
+                log_escape(msr->mp, col_name), log_escape_ex(msr->mp, col_key, col_key_len));
+    }
 
-    msr_log(msr, 4, "Retrieved collection (name \"%s\", key \"%s\").",
-            log_escape(msr->mp, col_name), log_escape(msr->mp, col_key));
+    if ((existing_dbm == NULL) && dbm) {
+        /* Should not ever get here */
+        msr_log(msr, 1, "Internal Error: Collection remained open (name \"%s\", key \"%s\").",
+                log_escape(msr->mp, col_name), log_escape_ex(msr->mp, col_key, col_key_len));
+
+        apr_sdbm_close(dbm);
+    }
 
     return col;
+
+cleanup:
+
+    if ((existing_dbm == NULL) && dbm) {
+        apr_sdbm_close(dbm);
+    }
+
+    return NULL;
 }
 
+/**
+ *
+ */
+apr_table_t *collection_retrieve(modsec_rec *msr, const char *col_name,
+    const char *col_key, int col_key_len) {
+    return collection_retrieve_ex(NULL, msr, col_name, col_key, col_key_len);
+}
+
+
 /**
  *
  */
@@ -193,84 +315,38 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
     apr_status_t rc;
     apr_sdbm_datum_t key;
     apr_sdbm_datum_t value;
-    apr_sdbm_t *dbm;
+    apr_sdbm_t *dbm = NULL;
     const apr_array_header_t *arr;
     apr_table_entry_t *te;
     int i;
+    const apr_table_t *stored_col = NULL;
+    const apr_table_t *orig_col = NULL;
 
     var_name = (msc_string *)apr_table_get(col, "__name");
     if (var_name == NULL) {
-        return -1;
+        goto error;
     }
 
     var_key = (msc_string *)apr_table_get(col, "__key");
     if (var_key == NULL) {
-        return -1;
+        goto error;
     }
 
     if (msr->txcfg->data_dir == NULL) {
         msr_log(msr, 1, "Unable to store collection (name \"%s\", key \"%s\"). Use "
             "SecDataDir to define data directory first.",
-            log_escape(msr->mp, var_name->value), log_escape(msr->mp, var_key->value));
-        return -1;
+            log_escape_ex(msr->mp, var_name->value, var_name->value_len), log_escape_ex(msr->mp, var_key->value, var_key->value_len));
+        goto error;
     }
 
-    dbm_filename = apr_pstrcat(msr->mp, msr->txcfg->data_dir, "/", var_name->value, NULL);    
+    // ENH: lowercase the var name in the filename
+    dbm_filename = apr_pstrcat(msr->mp, msr->txcfg->data_dir, "/", var_name->value, NULL);
 
-    /* Remove expired variables. */
-    do {
-        arr = apr_table_elts(col);
-        te = (apr_table_entry_t *)arr->elts;
-        for (i = 0; i < arr->nelts; i++) {
-            if (strncmp(te[i].key, "__expire_", 9) == 0) {
-                msc_string *var = (msc_string *)te[i].val;
-                int expiry_time = atoi(var->value);
+    /* Delete IS_NEW on store. */
+    apr_table_unset(col, "IS_NEW");
 
-                /* Do not remove the record itself. */
-                if (strcmp(te[i].key, "__expire_KEY") == 0) continue;                
-
-                if (expiry_time <= apr_time_sec(msr->request_time)) {
-                    char *key_to_expire = apr_pstrdup(msr->mp, te[i].key);
-                    msr_log(msr, 9, "Removing key \"%s\" from collection.", key_to_expire + 9);
-                    apr_table_unset(col, key_to_expire + 9);
-                    msr_log(msr, 9, "Removing key \"%s\" from collection.", key_to_expire);
-                    apr_table_unset(col, key_to_expire);
-                    msr_log(msr, 4, "Removed expired variable \"%s\".", key_to_expire + 9);
-                    break;
-                }
-            }
-        }
-    } while(i != arr->nelts);
-
-    /* Delete the collection if the variable "KEY" does not exist. */
-    if (apr_table_get(col, "KEY") == NULL) {
-
-        rc = apr_sdbm_open(&dbm, dbm_filename, APR_CREATE | APR_WRITE | APR_SHARELOCK,
-            CREATEMODE, msr->mp);
-        if (rc != APR_SUCCESS) {
-            msr_log(msr, 1, "Failed to access DBM file \"%s\": %s",
-                log_escape(msr->mp, dbm_filename), get_apr_error(msr->mp, rc));
-            return -1;
-        }
-
-        key.dptr = var_key->value;
-        key.dsize = var_key->value_len + 1;
-
-        rc = apr_sdbm_delete(dbm, key);
-        if (rc != APR_SUCCESS) {
-            msr_log(msr, 1, "Failed deleting collection (name \"%s\", "
-                "key \"%s\"): %s", log_escape(msr->mp, var_name->value),
-                log_escape(msr->mp, var_key->value), get_apr_error(msr->mp, rc));
-            apr_sdbm_close(dbm);
-            return -1;
-        }
-
-        msr_log(msr, 4, "Deleted collection (name \"%s\", key \"%s\").", 
-            log_escape(msr->mp, var_name->value), log_escape(msr->mp, var_key->value));
-        apr_sdbm_close(dbm);
-
-        return 1;
-    }
+    /* Delete UPDATE_RATE on store to save space as it is calculated */
+    apr_table_unset(col, "UPDATE_RATE");
 
     /* Update the timeout value. */
     {
@@ -279,7 +355,7 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
             int timeout = atoi(var->value);
             var = (msc_string *)apr_table_get(col, "__expire_KEY");
             if (var != NULL) {
-                var->value = apr_psprintf(msr->mp, "%i", (int)(apr_time_sec(apr_time_now()) + timeout));
+                var->value = apr_psprintf(msr->mp, "%" APR_TIME_T_FMT, (apr_time_t)(apr_time_sec(apr_time_now()) + timeout));
                 var->value_len = strlen(var->value);
             }
         }
@@ -294,7 +370,7 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
             var->name_len = strlen(var->name);
             apr_table_setn(col, var->name, (void *)var);
         }
-        var->value = apr_psprintf(msr->mp, "%i", (int)(apr_time_sec(apr_time_now())));
+        var->value = apr_psprintf(msr->mp, "%" APR_TIME_T_FMT, (apr_time_t)(apr_time_sec(apr_time_now())));
         var->value_len = strlen(var->value);
     }
 
@@ -310,11 +386,44 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
         } else {
             counter = atoi(var->value);
         }
-        var->value = apr_psprintf(msr->mp, "%i", counter + 1);
+        var->value = apr_psprintf(msr->mp, "%d", counter + 1);
         var->value_len = strlen(var->value);
     }
 
-    /* Calculate the size first. */
+    /* ENH Make the expiration timestamp accessible in blob form so that
+     * it is easier/faster to determine expiration without having to
+     * convert back to table form
+     */
+
+    rc = apr_sdbm_open(&dbm, dbm_filename, APR_CREATE | APR_WRITE | APR_SHARELOCK,
+        CREATEMODE, msr->mp);
+    if (rc != APR_SUCCESS) {
+        msr_log(msr, 1, "Failed to access DBM file \"%s\": %s", log_escape(msr->mp, dbm_filename),
+            get_apr_error(msr->mp, rc));
+        dbm = NULL;
+        goto error;
+    }
+
+    /* Need to lock to pull in the stored data again and apply deltas. */
+    rc = apr_sdbm_lock(dbm, APR_FLOCK_EXCLUSIVE);
+    if (rc != APR_SUCCESS) {
+        msr_log(msr, 1, "Failed to exclusivly lock DBM file \"%s\": %s", log_escape(msr->mp, dbm_filename),
+            get_apr_error(msr->mp, rc));
+        goto error;
+    }
+
+    /* If there is an original value, then create a delta and
+     * apply the delta to the current value */
+    orig_col = (const apr_table_t *)apr_table_get(msr->collections_original, var_name->value);
+    if (orig_col != NULL) {
+        if (msr->txcfg->debuglog_level >= 9) {
+            msr_log(msr, 9, "Re-retrieving collection prior to store: %s", apr_psprintf(msr->mp, "%.*s", var_name->value_len, var_name->value));
+        }
+
+        stored_col = (const apr_table_t *)collection_retrieve_ex(dbm, msr, var_name->value, var_key->value, var_key->value_len);
+    }
+
+    /* Merge deltas and calculate the size first. */
     blob_size = 3 + 2;
     arr = apr_table_elts(col);
     te = (apr_table_entry_t *)arr->elts;
@@ -322,6 +431,33 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
         msc_string *var = (msc_string *)te[i].val;
         int len;
 
+        /* If there is an original value, then apply the delta
+         * to the latest stored value */
+        if (stored_col != NULL) {
+            const msc_string *orig_var = (const msc_string *)apr_table_get(orig_col, var->name);
+            if (orig_var != NULL) {
+                const msc_string *stored_var = (const msc_string *)apr_table_get(stored_col, var->name);
+                if (stored_var != NULL) {
+                    int origval = atoi(orig_var->value);
+                    int ourval = atoi(var->value);
+                    int storedval = atoi(stored_var->value);
+                    int delta = ourval - origval;
+                    int newval = storedval + delta;
+
+                    if (newval < 0) newval = 0; /* Counters never go below zero. */
+
+                    var->value = apr_psprintf(msr->mp, "%d", newval);
+                    var->value_len = strlen(var->value);
+                    if (msr->txcfg->debuglog_level >= 9) {
+                        msr_log(msr, 9, "Delta applied for %s.%s %d->%d (%d): %d + (%d) = %d [%s,%d]",
+                        log_escape_ex(msr->mp, var_name->value, var_name->value_len),
+                        log_escape_ex(msr->mp, var->name, var->name_len),
+                        origval, ourval, delta, storedval, delta, newval, var->value, var->value_len);
+                    }
+                }
+            }
+        }
+
         len = var->name_len + 1;
         if (len >= 65536) len = 65536;
         blob_size += len + 2;
@@ -334,7 +470,7 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
     /* Now generate the binary object. */
     blob = apr_pcalloc(msr->mp, blob_size);
     if (blob == NULL) {
-        return -1;
+        goto error;
     }
 
     blob[0] = 0x49;
@@ -366,25 +502,17 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
         blob[blob_offset + 2 + len - 1] = '\0';
         blob_offset += 2 + len;
 
-        msr_log(msr, 9, "Wrote variable: name \"%s\", value \"%s\".",
-            log_escape_ex(msr->mp, var->name, var->name_len),
-            log_escape_ex(msr->mp, var->value, var->value_len));
+        if (msr->txcfg->debuglog_level >= 9) {
+            msr_log(msr, 9, "Wrote variable: name \"%s\", value \"%s\".",
+                log_escape_ex(msr->mp, var->name, var->name_len),
+                log_escape_ex(msr->mp, var->value, var->value_len));
+        }
     }
 
     blob[blob_offset] = 0;
     blob[blob_offset + 1] = 0;
 
     /* And, finally, store it. */
-    dbm_filename = apr_pstrcat(msr->mp, msr->txcfg->data_dir, "/", var_name->value, NULL);    
-
-    rc = apr_sdbm_open(&dbm, dbm_filename, APR_CREATE | APR_WRITE | APR_SHARELOCK,
-        CREATEMODE, msr->mp);
-    if (rc != APR_SUCCESS) {
-        msr_log(msr, 1, "Failed to access DBM file \"%s\": %s", log_escape(msr->mp, dbm_filename),
-            get_apr_error(msr->mp, rc));
-        return -1;
-    }
-
     key.dptr = var_key->value;
     key.dsize = var_key->value_len + 1;
 
@@ -395,16 +523,25 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
     if (rc != APR_SUCCESS) {
         msr_log(msr, 1, "Failed to write to DBM file \"%s\": %s", dbm_filename,
             get_apr_error(msr->mp, rc));
-        apr_sdbm_close(dbm);
-        return -1;
-    }    
-
-    msr_log(msr, 4, "Persisted collection (name \"%s\", key \"%s\").",
-        log_escape(msr->mp, var_name->value), log_escape(msr->mp, var_key->value));
+        goto error;
+    }
 
     apr_sdbm_close(dbm);
 
-    return 0;    
+    if (msr->txcfg->debuglog_level >= 4) {
+        msr_log(msr, 4, "Persisted collection (name \"%s\", key \"%s\").",
+            log_escape_ex(msr->mp, var_name->value, var_name->value_len), log_escape_ex(msr->mp, var_key->value, var_key->value_len));
+    }
+
+    return 0;
+
+error:
+
+    if (dbm) {
+        apr_sdbm_close(dbm);
+    }
+
+    return -1;
 }
 
 /**
@@ -413,19 +550,19 @@ int collection_store(modsec_rec *msr, apr_table_t *col) {
 int collections_remove_stale(modsec_rec *msr, const char *col_name) {
     char *dbm_filename = NULL;
     apr_sdbm_datum_t key, value;
-    apr_sdbm_t *dbm;
+    apr_sdbm_t *dbm = NULL;
     apr_status_t rc;
     apr_array_header_t *keys_arr;
     char **keys;
+    apr_time_t now = apr_time_sec(msr->request_time);
     int i;
-    unsigned int now = (unsigned int)apr_time_sec(msr->request_time);
 
     if (msr->txcfg->data_dir == NULL) {
-        /* The user has been warned about this problem enough times already by now. 
+        /* The user has been warned about this problem enough times already by now.
          * msr_log(msr, 1, "Unable to access collection file (name \"%s\"). Use SecDataDir to "
          *     "define data directory first.", log_escape(msr->mp, col_name));
          */
-        return -1;
+        goto error;
     }
 
     dbm_filename = apr_pstrcat(msr->mp, msr->txcfg->data_dir, "/", col_name, NULL);
@@ -435,7 +572,8 @@ int collections_remove_stale(modsec_rec *msr, const char *col_name) {
     if (rc != APR_SUCCESS) {
         msr_log(msr, 1, "Failed to access DBM file \"%s\": %s", log_escape(msr->mp, dbm_filename),
             get_apr_error(msr->mp, rc));
-        return -1;
+        dbm = NULL;
+        goto error;
     }
 
     /* First get a list of all keys. */
@@ -444,23 +582,24 @@ int collections_remove_stale(modsec_rec *msr, const char *col_name) {
     if (rc != APR_SUCCESS) {
         msr_log(msr, 1, "Failed to lock DBM file \"%s\": %s", log_escape(msr->mp, dbm_filename),
             get_apr_error(msr->mp, rc));
-        apr_sdbm_close(dbm);
-        return -1;
+        goto error;
     }
 
-    /* No one can write to the file while we're
-     * doing this so let's do it as fast as we can.
+    /* No one can write to the file while doing this so
+     * do it as fast as possible.
      */
     rc = apr_sdbm_firstkey(dbm, &key);
     while(rc == APR_SUCCESS) {
-        char *s = apr_pstrmemdup(msr->mp, key.dptr, key.dsize);
+        char *s = apr_pstrmemdup(msr->mp, key.dptr, key.dsize - 1);
         *(char **)apr_array_push(keys_arr) = s;
         rc = apr_sdbm_nextkey(dbm, &key);
     }
     apr_sdbm_unlock(dbm);
 
-    msr_log(msr, 9, "Found %i record(s) in file \"%s\".", keys_arr->nelts,
-        log_escape(msr->mp, dbm_filename));
+    if (msr->txcfg->debuglog_level >= 9) {
+        msr_log(msr, 9, "Found %d record(s) in file \"%s\".", keys_arr->nelts,
+            log_escape(msr->mp, dbm_filename));
+    }
 
     /* Now retrieve the entires one by one. */
     keys = (char **)keys_arr->elts;
@@ -472,42 +611,45 @@ int collections_remove_stale(modsec_rec *msr, const char *col_name) {
         if (rc != APR_SUCCESS) {
             msr_log(msr, 1, "Failed reading DBM file \"%s\": %s",
                 log_escape(msr->mp, dbm_filename), get_apr_error(msr->mp, rc));
-            apr_sdbm_close(dbm);
-            return -1;
+            goto error;
         }
 
         if (value.dptr != NULL) {
             apr_table_t *col = NULL;
             msc_string *var = NULL;
 
-            col = collection_unpack(msr, value.dptr, value.dsize, 0);
+            col = collection_unpack(msr, (const unsigned char *)value.dptr, value.dsize, 0);
             if (col == NULL) {
-                return -1;
+                goto error;
             }
 
             var = (msc_string *)apr_table_get(col, "__expire_KEY");
             if (var == NULL) {
                 msr_log(msr, 1, "Collection cleanup discovered entry with no "
                     "__expire_KEY (name \"%s\", key \"%s\").",
-                    log_escape(msr->mp, col_name), log_escape(msr->mp, key.dptr));
+                    log_escape(msr->mp, col_name), log_escape_ex(msr->mp, key.dptr, key.dsize - 1));
             } else {
                 unsigned int expiry_time = atoi(var->value);
 
-                msr_log(msr, 9, "Record (name \"%s\", key \"%s\") set to expire in %i seconds.",
-                    log_escape(msr->mp, col_name), log_escape(msr->mp, key.dptr),
-                    expiry_time - now);
+                if (msr->txcfg->debuglog_level >= 9) {
+                    msr_log(msr, 9, "Record (name \"%s\", key \"%s\") set to expire in %" APR_TIME_T_FMT " seconds.",
+                        log_escape(msr->mp, col_name), log_escape_ex(msr->mp, key.dptr, key.dsize - 1),
+                        expiry_time - now);
+                }
 
                 if (expiry_time <= now) {
                     rc = apr_sdbm_delete(dbm, key);
                     if (rc != APR_SUCCESS) {
                         msr_log(msr, 1, "Failed deleting collection (name \"%s\", "
                             "key \"%s\"): %s", log_escape(msr->mp, col_name),
-                            log_escape(msr->mp, key.dptr), get_apr_error(msr->mp, rc));
-                        return -1;
+                            log_escape_ex(msr->mp, key.dptr, key.dsize - 1), get_apr_error(msr->mp, rc));
+                        goto error;
+                    }
+                    if (msr->txcfg->debuglog_level >= 4) {
+                        msr_log(msr, 4, "Removed stale collection (name \"%s\", "
+                                "key \"%s\").", log_escape(msr->mp, col_name),
+                                log_escape_ex(msr->mp, key.dptr, key.dsize - 1));
                     }
-                    msr_log(msr, 4, "Removed stale collection (name \"%s\", "
-                            "key \"%s\").", log_escape(msr->mp, col_name),
-                            log_escape(msr->mp, key.dptr));
                 }
             }
         } else {
@@ -518,4 +660,12 @@ int collections_remove_stale(modsec_rec *msr, const char *col_name) {
     apr_sdbm_close(dbm);
 
     return 1;
+
+error:
+
+    if (dbm) {
+        apr_sdbm_close(dbm);
+    }
+
+    return -1;
 }
